from django.conf.urls import patterns, url
from django.contrib.admin.views.decorators import staff_member_required

from oscar.core.application import Application

from . import views


class DatacashDashboardApplication(Application):
    name = None
    list_view = views.TransactionListView
    detail_view = views.TransactionDetailView
    fraud_list_view = views.FraudResponseListView

    def get_urls(self):
        urlpatterns = patterns('',
            url(r'^transactions/$', self.list_view.as_view(),
                name='datacash-transaction-list'),
            url(r'^transactions/(?P<pk>\d+)/$', self.detail_view.as_view(),
                name='datacash-transaction-detail'),
            url(r'^fraud-responses/$', self.fraud_list_view.as_view(),
                name='datacash-fraud-response-list'),
        )
        return self.post_process_urls(urlpatterns)

    def get_url_decorator(self, url_name):
        return staff_member_required


application = DatacashDashboardApplication()



from django.conf.urls import *
from django.views.decorators.csrf import csrf_exempt

from .the3rdman import views


# Responses from the3rdman are posted back
urlpatterns = patterns('',
    url(r'^the3rdman/', csrf_exempt(views.CallbackView.as_view()),
        name='datacash-3rdman-callback'),
)



from django.db.models import get_model

Order = get_model('order', 'Order')


def build_data_dict(request=None, user=None, email=None, order_number=None,
                    basket=None, shipping_address=None, billing_address=None):
    """
    Build a dict of the fields that can be passed as 3rdMan data
    """
    if request and user is None:
        user = request.user
    if request and basket is None:
        basket = request.basket
    data = {
        'customer_info': build_customer_info(
            request, user, email, order_number, shipping_address),
        'delivery_info': build_delivery_info(
            shipping_address),
        'billing_info': build_billing_info(
            billing_address),
        'account_info': {},  # Not implemented for now
        'order_info': build_order_info(basket),
    }
    return data


def build_customer_info(request, user, email, order_number, shipping_address):
    # We only use fields denoted 'R' (for Retail)
    payload = {}

    if user and user.is_authenticated():
        payload['customer_reference'] = user.id
        payload['email'] = user.email
        payload['forename'] = user.first_name
        payload['surname'] = user.last_name

    if email:
        payload['email'] = email

    if shipping_address:
        payload['delivery_forename'] = shipping_address.first_name
        payload['delivery_phone_number'] = shipping_address.phone_number
        payload['delivery_surname'] = shipping_address.last_name
        payload['delivery_title'] = shipping_address.title

    if request and 'REMOTE_ADDR' in request.META:
        payload['ip_address'] = request.META['REMOTE_ADDR']
    # We let HTTP_X_FORWARDED_FOR take precedence if it exists
    if request and 'HTTP_X_FORWARDED_FOR' in request.META:
        payload['ip_address'] = request.META['HTTP_X_FORWARDED_FOR']

    if order_number:
        payload['order_number'] = order_number

    payload['sales_channel'] = 3  # Internet

    return payload


def build_delivery_info(shipping_address):
    payload = {}
    if not shipping_address:
        return payload
    payload['street_address_1'] = shipping_address.line1
    payload['street_address_2'] = shipping_address.line2
    payload['city'] = shipping_address.line4
    payload['county'] = shipping_address.state
    payload['postcode'] = shipping_address.postcode
    payload['country'] = u"%.03d" % shipping_address.country.iso_3166_1_numeric
    return payload


def build_billing_info(billing_address):
    payload = {}
    if not billing_address:
        return payload
    payload['street_address_1'] = billing_address.line1
    payload['street_address_2'] = billing_address.line2
    payload['city'] = billing_address.line4
    payload['county'] = billing_address.state
    payload['postcode'] = billing_address.postcode
    return payload


def build_order_info(basket):
    if not basket:
        return {}
    payload = {'products': []}
    for line in basket.all_lines():
        product = line.product
        stockrecord = product.stockrecord
        datum = {
            'code': product.upc,
            'price': stockrecord.price_incl_tax,
            'prod_description': product.description,
            'prod_id': product.id,
            'quantity': line.quantity,
        }
        payload['products'].append(datum)
    return payload



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Removing unique constraint on 'FraudResponse', fields ['t3m_id']
        db.delete_unique('datacash_fraudresponse', ['t3m_id'])

        # Adding index on 'FraudResponse', fields ['t3m_id']
        db.create_index('datacash_fraudresponse', ['t3m_id'])


    def backwards(self, orm):
        # Removing index on 'FraudResponse', fields ['t3m_id']
        db.delete_index('datacash_fraudresponse', ['t3m_id'])

        # Adding unique constraint on 'FraudResponse', fields ['t3m_id']
        db.create_unique('datacash_fraudresponse', ['t3m_id'])


    models = {
        'datacash.fraudresponse': {
            'Meta': {'ordering': "('-date_created',)", 'object_name': 'FraudResponse'},
            'aggregator_identifier': ('django.db.models.fields.CharField', [], {'max_length': '15', 'blank': 'True'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'merchant_identifier': ('django.db.models.fields.CharField', [], {'max_length': '15'}),
            'merchant_order_ref': ('django.db.models.fields.CharField', [], {'max_length': '250', 'db_index': 'True'}),
            'message_digest': ('django.db.models.fields.CharField', [], {'max_length': '128', 'blank': 'True'}),
            'raw_response': ('django.db.models.fields.TextField', [], {}),
            'recommendation': ('django.db.models.fields.IntegerField', [], {}),
            'score': ('django.db.models.fields.IntegerField', [], {}),
            't3m_id': ('django.db.models.fields.CharField', [], {'max_length': '128', 'db_index': 'True'})
        },
        'datacash.ordertransaction': {
            'Meta': {'ordering': "('-date_created',)", 'object_name': 'OrderTransaction'},
            'amount': ('django.db.models.fields.DecimalField', [], {'null': 'True', 'max_digits': '12', 'decimal_places': '2', 'blank': 'True'}),
            'auth_code': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'datacash_reference': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'merchant_reference': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'method': ('django.db.models.fields.CharField', [], {'max_length': '12'}),
            'order_number': ('django.db.models.fields.CharField', [], {'max_length': '128', 'db_index': 'True'}),
            'reason': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'request_xml': ('django.db.models.fields.TextField', [], {}),
            'response_xml': ('django.db.models.fields.TextField', [], {}),
            'status': ('django.db.models.fields.PositiveIntegerField', [], {})
        }
    }

    complete_apps = ['datacash']


from datacash.xmlutils import create_element
from xml.dom.minidom import Document, parseString


def add_fraud_fields(doc=None, element=None, customer_info=None, delivery_info=None,
                     billing_info=None, account_info=None, order_info=None,
                     **kwargs):
    """
    Submit a transaction to The3rdMan for Batch Fraud processing

    Doesn't support (yet):
        - Previous orders
        - Additional delivery addreses
        - Alternative payments

    See section 2.4.7 of the Developer Guide
    """
    # Build the request XML
    if doc is None:
        doc = Document()
    if element is None:
        element = doc

    envelope = create_element(
        doc, element, 'The3rdMan', attributes={'type': 'realtime'})

    if 'callback_url' in kwargs:
        callback_format = kwargs.get('callback_format', 'XML')
        callback_url = kwargs['callback_url']
        add_realtime_information(doc, envelope, callback_format, callback_url)

    add_customer_information(doc, envelope, customer_info)
    add_delivery_address(doc, envelope, delivery_info)
    add_billing_address(doc, envelope, billing_info)
    add_account_information(doc, envelope, account_info)
    add_order_information(doc, envelope, order_info)
    return doc


def add_realtime_information(doc, ele, format, url):
    rt_ele = create_element(doc, ele, 'Realtime')
    create_element(doc, rt_ele, 'real_time_callback_format', format)
    create_element(doc, rt_ele, 'real_time_callback', url)


def add_xml_fields(doc, parent, fields, values):
    for field in fields:
        if field in values and values[field] is not None:
            create_element(doc, parent, field, values[field])


def intersects(fields, values):
    """
    Test if there are any fields in the values
    """
    overlap = set(fields).intersection(values.keys())
    return len(overlap) > 0


def add_customer_information(doc, envelope, customer_info):
    if not customer_info:
        return
    cust_ele = create_element(doc, envelope, 'CustomerInformation')
    cust_fields = (
        'alt_telephone', 'customer_dob', 'customer_reference',
        'delivery_forename', 'delivery_phone_number', 'delivery_surname',
        'delivery_title', 'driving_license_number', 'email',
        'first_purchase_date', 'forename', 'introduced_by', 'ip_address',
        'order_number', 'sales_channel', 'surname', 'telephone',
        'time_zone', 'title')
    add_xml_fields(doc, cust_ele, cust_fields, customer_info)


def add_delivery_address(doc, envelope, delivery_info):
    if not delivery_info:
        return
    delivery_ele = create_element(doc, envelope, 'DeliveryAddress')
    delivery_fields = (
        'street_address_1', 'street_address_2', 'city',
        'county', 'postcode', 'country')
    add_xml_fields(doc, delivery_ele, delivery_fields, delivery_info)


def add_billing_address(doc, envelope, billing_info):
    if not billing_info:
        return
    billing_ele = create_element(doc, envelope, 'BillingAddress')
    billing_fields = (
        'street_address_1', 'street_address_2', 'city', 'county',
        'postcode', 'country')
    add_xml_fields(doc, billing_ele, billing_fields, billing_info)


def add_account_information(doc, envelope, account_info):
    if not account_info:
        return
    account_ele = create_element(doc, envelope, 'AccountInformation')

    # Bank information
    bank_fields = (
        'account_number', 'bank_address', 'bank_country', 'bank_name',
        'customer_name', 'sort_code')
    if intersects(bank_fields, account_info):
        bank_ele = create_element(doc, account_ele, 'BankInformation')
        add_xml_fields(doc, bank_ele, bank_fields, account_info)

    purchase_fields = ('avg', 'max', 'min')
    if intersects(purchase_fields, account_info):
        purchase_ele = create_element(doc, account_ele, 'PurchaseInformation')
        add_xml_fields(doc, purchase_ele, purchase_fields, account_info)


def add_order_information(doc, envelope, order_info):
    if not order_info or 'products' not in order_info:
        return
    order_ele = create_element(doc, envelope, 'OrderInformation')
    products_ele = create_element(
        doc, order_ele, 'Products',
        attributes={'count': len(order_info['products'])})
    product_fields = ('code', 'prod_id', 'quantity', 'price', 'prod_category',
                      'prod_description')
    for product_info in order_info['products']:
        product_ele = create_element(doc, products_ele, 'Product')
        add_xml_fields(doc, product_ele, product_fields, product_info)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'FraudResponse'
        db.create_table('datacash_fraudresponse', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('aggregator_identifier', self.gf('django.db.models.fields.CharField')(max_length=15, blank=True)),
            ('merchant_identifier', self.gf('django.db.models.fields.CharField')(max_length=15)),
            ('merchant_order_ref', self.gf('django.db.models.fields.CharField')(max_length=250, db_index=True)),
            ('t3m_id', self.gf('django.db.models.fields.CharField')(unique=True, max_length=128)),
            ('score', self.gf('django.db.models.fields.IntegerField')()),
            ('recommendation', self.gf('django.db.models.fields.IntegerField')()),
            ('message_digest', self.gf('django.db.models.fields.CharField')(max_length=128, blank=True)),
            ('raw_response', self.gf('django.db.models.fields.TextField')()),
            ('date_created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
        ))
        db.send_create_signal('datacash', ['FraudResponse'])


    def backwards(self, orm):
        # Deleting model 'FraudResponse'
        db.delete_table('datacash_fraudresponse')


    models = {
        'datacash.fraudresponse': {
            'Meta': {'object_name': 'FraudResponse'},
            'aggregator_identifier': ('django.db.models.fields.CharField', [], {'max_length': '15', 'blank': 'True'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'merchant_identifier': ('django.db.models.fields.CharField', [], {'max_length': '15'}),
            'merchant_order_ref': ('django.db.models.fields.CharField', [], {'max_length': '250', 'db_index': 'True'}),
            'message_digest': ('django.db.models.fields.CharField', [], {'max_length': '128', 'blank': 'True'}),
            'raw_response': ('django.db.models.fields.TextField', [], {}),
            'recommendation': ('django.db.models.fields.IntegerField', [], {}),
            'score': ('django.db.models.fields.IntegerField', [], {}),
            't3m_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '128'})
        },
        'datacash.ordertransaction': {
            'Meta': {'ordering': "('-date_created',)", 'object_name': 'OrderTransaction'},
            'amount': ('django.db.models.fields.DecimalField', [], {'null': 'True', 'max_digits': '12', 'decimal_places': '2', 'blank': 'True'}),
            'auth_code': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'datacash_reference': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'merchant_reference': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'method': ('django.db.models.fields.CharField', [], {'max_length': '12'}),
            'order_number': ('django.db.models.fields.CharField', [], {'max_length': '128', 'db_index': 'True'}),
            'reason': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'request_xml': ('django.db.models.fields.TextField', [], {}),
            'response_xml': ('django.db.models.fields.TextField', [], {}),
            'status': ('django.db.models.fields.PositiveIntegerField', [], {})
        }
    }

    complete_apps = ['datacash']


def create_element(doc, parent, tag, value=None, attributes=None):
    """
    Creates an XML element
    """
    ele = doc.createElement(tag)
    parent.appendChild(ele)
    if value:
        text = doc.createTextNode(u"%s" % value)
        ele.appendChild(text)
    if attributes:
        [ele.setAttribute(k, str(v)) for k, v in attributes.items()]
    return ele



from django.views.generic import ListView, DetailView

from datacash import models


class TransactionListView(ListView):
    model = models.OrderTransaction
    context_object_name = 'transactions'
    template_name = 'datacash/dashboard/transaction_list.html'


class TransactionDetailView(DetailView):
    model = models.OrderTransaction
    context_object_name = 'txn'
    template_name = 'datacash/dashboard/transaction_detail.html'


class FraudResponseListView(ListView):
    model = models.FraudResponse
    context_object_name = 'responses'
    template_name = 'datacash/dashboard/fraudresponse_list.html'



import re
from xml.dom.minidom import parseString
import urlparse

from django.db import models
from django.conf import settings

from .the3rdman import signals


def prettify_xml(xml_str):
    xml_str = re.sub(r'\s*\n\s*', '', xml_str)
    ugly = parseString(xml_str.encode('utf8')).toprettyxml(indent='    ')
    regex = re.compile(r'>\n\s+([^<>\s].*?)\n\s+</', re.DOTALL)
    return regex.sub('>\g<1></', ugly)


class OrderTransaction(models.Model):

    # Note we don't use a foreign key as the order hasn't been created
    # by the time the transaction takes place
    order_number = models.CharField(max_length=128, db_index=True)

    # The 'method' of the transaction - one of 'auth', 'pre', 'cancel', ...
    method = models.CharField(max_length=12)
    amount = models.DecimalField(
        decimal_places=2, max_digits=12, blank=True, null=True)
    currency = models.CharField(
        max_length=12, default=settings.DATACASH_CURRENCY)
    merchant_reference = models.CharField(
        max_length=128, blank=True, null=True)

    # Response fields
    datacash_reference = models.CharField(max_length=128, blank=True, null=True)
    auth_code = models.CharField(max_length=128, blank=True, null=True)
    status = models.PositiveIntegerField()
    reason = models.CharField(max_length=255)

    # Store full XML for debugging purposes
    request_xml = models.TextField()
    response_xml = models.TextField()

    date_created = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('-date_created',)

    def _replace_credit_card_number(self, matchobj):
        """ Credit card number can be from 13 to 19 digits long. Shown only
        last 4 of them and replace others with 'X' still keeping
        number of digits """
        return "<%(element)s>%(hidden)s%(last4)s</%(element)s>" % {
            'element': matchobj.group(1),
            'hidden': "X" * len(matchobj.group(2)),
            'last4': matchobj.group(3),
        }

    def save(self, *args, **kwargs):
        # Ensure sensitive data isn't saved
        if not self.pk:
            cc_regex = re.compile(r'<(pan|alt_pan)>(\d+)(\d{4})</\1>')
            self.request_xml = cc_regex.sub(self._replace_credit_card_number,
                                            self.request_xml)
            ccv_regex = re.compile(r'<cv2>\d+</cv2>')
            self.request_xml = ccv_regex.sub('<cv2>XXX</cv2>', self.request_xml)
            pw_regex = re.compile(r'<password>.*</password>')
            self.request_xml = pw_regex.sub('<password>XXX</password>', self.request_xml)
        super(OrderTransaction, self).save(*args, **kwargs)

    def __unicode__(self):
        return u'%s txn for order %s - ref: %s, status: %s' % (
            self.method.upper(),
            self.order_number,
            self.datacash_reference,
            self.status)

    @property
    def pretty_request_xml(self):
        return prettify_xml(self.request_xml)

    @property
    def pretty_response_xml(self):
        return prettify_xml(self.response_xml)

    @property
    def accepted(self):
        return self.status == 1

    @property
    def declined(self):
        return self.status == 7


class FraudResponse(models.Model):
    aggregator_identifier = models.CharField(max_length=15, blank=True)
    merchant_identifier = models.CharField(max_length=15)
    merchant_order_ref = models.CharField(max_length=250, db_index=True)
    t3m_id = models.CharField(max_length=128, db_index=True)
    score = models.IntegerField()

    RELEASE, HOLD, REJECT, UNDER_INVESTIGATION = 0, 1, 2, 9
    recommendation = models.IntegerField()
    message_digest = models.CharField(max_length=128, blank=True)
    raw_response = models.TextField()
    date_created = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        return u"t3m ID %s (score: %s, recommendation: %s)" % (
            self.t3m_id, self.score, self.recommendation)

    class Meta:
        ordering = ('-date_created',)

    @classmethod
    def create_from_xml(cls, xml_string):
        """
        Create a fraud response instance from an XML payload
        """
        # Helper function for text extraction
        def tag_text(doc, tag_name):
            try:
                ele = doc.getElementsByTagName(tag_name)[0]
            except IndexError:
                return ''
            if ele.firstChild:
                return ele.firstChild.data
            return ''

        doc = parseString(xml_string)
        return cls.create_from_payload(xml_string, doc, tag_text)

    @classmethod
    def create_from_querystring(cls, query):
        """
        Create a fraud response instance from a querystring payload
        """
        def extract(data, key):
            return data.get(key, [""])[0]

        data = urlparse.parse_qs(query)
        return cls.create_from_payload(query, data, extract)

    @classmethod
    def create_from_payload(cls, raw, payload, extract_fn):
        response = cls.objects.create(
            aggregator_identifier=extract_fn(payload, 'aggregator_identifier'),
            merchant_identifier=extract_fn(payload, 'merchant_identifier'),
            merchant_order_ref=extract_fn(payload, 'merchant_order_ref'),
            t3m_id=extract_fn(payload, 't3m_id'),
            score=int(extract_fn(payload, 'score')),
            recommendation=int(extract_fn(payload, 'recommendation')),
            message_digest=extract_fn(payload, 'message_digest'),
            raw_response=raw)

        # Raise signal so other processes can update orders based on this fraud
        # response.
        signals.response_received.send_robust(sender=cls, response=response)

        return response

    @property
    def on_hold(self):
        return self.recommendation == self.HOLD

    @property
    def released(self):
        return self.recommendation == self.RELEASE

    @property
    def rejected(self):
        return self.recommendation == self.REJECT

    @property
    def order_number(self):
        """
        Return the order number from the original transaction.

        This assumes the merchant ref was generated using the datacash.facade
        class
        """
        return self.merchant_order_ref.split("_")[0]

    @property
    def recommendation_text(self):
        mapping = {
            self.RELEASE: "Released",
            self.HOLD: "On hold",
            self.REJECT: "Rejected",
            self.UNDER_INVESTIGATION: "Under investigation",
        }
        return mapping.get(self.recommendation, "Unknown")

    @property
    def gatekeeper_url(self):
        """
        Return the transaction detail URL on the Gatekeeper site
        """
        is_live = 'mars' in settings.DATACASH_HOST
        host = 'cnpanalyst.com' if is_live else 'test.cnpanalyst.com'
        return 'https://%s/TransactionDetails.aspx?TID=%s' % (
            host, self.t3m_id)



import random

from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from oscar.apps.payment.exceptions import UnableToTakePayment, InvalidGatewayRequestError

from datacash import gateway
from datacash.models import OrderTransaction


class Facade(object):
    """
    A bridge between oscar's objects and the core gateway object
    """

    def __init__(self):
        self.gateway = gateway.Gateway(
            settings.DATACASH_HOST,
            getattr(settings, 'DATACASH_PATH', '/Transaction'),
            settings.DATACASH_CLIENT,
            settings.DATACASH_PASSWORD,
            getattr(settings, 'DATACASH_USE_CV2AVS', False),
            getattr(settings, 'DATACASH_CAPTURE_METHOD', 'ecomm'))

    def handle_response(self, method, order_number, amount, currency, response):

        # Maintain audit trail
        self.record_txn(method, order_number, amount, currency, response)

        # A response is either successful, declined or an error
        if response.is_successful():
            return response['datacash_reference']
        elif response.is_declined():
            msg = self.get_friendly_decline_message(response)
            raise UnableToTakePayment(msg)
        else:
            # If there's a customer friendly version of this error, raise
            # UnableToTakePayment so that the error gets shown to the customer,
            # otherwise raise InvalidGatewayRequestError (a subclass of
            # PaymentError) which will show a generic error message.
            friendly_msg = self.get_friendly_error_message(response)
            if friendly_msg:
                raise UnableToTakePayment(friendly_msg)
            raise InvalidGatewayRequestError(response.reason)

    def record_txn(self, method, order_number, amount, currency, response):
        OrderTransaction.objects.create(
            order_number=order_number,
            method=method,
            datacash_reference=response['datacash_reference'],
            merchant_reference=response['merchant_reference'],
            amount=amount,
            currency=currency or "",
            auth_code=response['auth_code'],
            status=response.status,
            reason=response['reason'],
            request_xml=response.request_xml,
            response_xml=response.response_xml)

    def get_friendly_decline_message(self, response):
        return _('The transaction was declined by your bank - '
                 'please check your bankcard details and try again')

    def get_friendly_error_message(self, response):
        # TODO: expand this dict to handle the most common errors
        errors = {
            56: _('This transaction was submitted too soon after the '
                  'previous one.  Please wait for a minute then try again'),
            59: _("We currently don't support your bankcard type.  If "
                  "possible, please use a different bankcard to try again"),
            19: _('Unable to fulfill transaction'),
        }
        return errors.get(response.status)

    def extract_address_data(self, address):
        data = {}
        if not address:
            return data
        for i in range(1, 5):
            key = 'line%d' % i
            if hasattr(address, key):
                data['address_line%d' % i] = getattr(address, key)
        data['postcode'] = address.postcode
        return data

    # ========================
    # API - 2 stage processing
    # ========================

    def pre_authorise(self, order_number, amount, bankcard=None,
                      txn_reference=None, billing_address=None,
                      the3rdman_data=None, currency=None):
        """
        Ring-fence an amount of money from the given card.  This is the first
        stage of a two-stage payment process.  A further call to fulfill is
        required to debit the money.

        As there are SO MANY values that can be submitted to 3rdMan, a separate
        dict of data must be submitted as a kwarg.
        """
        if amount == 0:
            raise UnableToTakePayment("Order amount must be non-zero")
        if currency is None:
            currency = settings.DATACASH_CURRENCY
        merchant_ref = self.merchant_reference(order_number, gateway.PRE)
        address_data = self.extract_address_data(billing_address)
        if bankcard:
            response = self.gateway.pre(
                card_number=bankcard.card_number,
                expiry_date=bankcard.expiry_date,
                amount=amount,
                currency=currency,
                merchant_reference=merchant_ref,
                ccv=bankcard.ccv,
                the3rdman_data=the3rdman_data,
                **address_data)
        elif txn_reference:
            response = self.gateway.pre(
                amount=amount,
                currency=currency,
                merchant_reference=merchant_ref,
                previous_txn_reference=txn_reference,
                the3rdman_data=the3rdman_data,
                **address_data)
        else:
            raise ValueError(
                "You must specify either a bankcard or a previous txn reference")
        return self.handle_response(
            gateway.PRE, order_number, amount, currency, response)

    def merchant_reference(self, order_number, method):
        # Determine the previous number of these transactions.
        num_previous = OrderTransaction.objects.filter(
            order_number=order_number, method=method).count()
        # Get a random number to append to the end.  This solves the problem
        # where a previous request crashed out and didn't save a model
        # instance.  Hence we can get a clash of merchant references.
        rand = "%04.f" % (random.random() * 10000)
        return u'%s_%s_%d_%s' % (order_number, method.upper(), num_previous+1,
                                 rand)

    def fulfill_transaction(self, order_number, amount, txn_reference,
                            auth_code, currency=None):
        """
        Settle a previously ring-fenced transaction
        """
        if currency is None:
            currency = settings.DATACASH_CURRENCY
        # Split shipments require that fulfills after the first one must have a
        # different merchant reference to the original
        merchant_ref = self.merchant_reference(order_number, gateway.FULFILL)
        response = self.gateway.fulfill(amount=amount,
                                        currency=currency,
                                        merchant_reference=merchant_ref,
                                        txn_reference=txn_reference,
                                        auth_code=auth_code)
        return self.handle_response(gateway.FULFILL, order_number, amount,
                                    currency, response)

    def refund_transaction(self, order_number, amount, txn_reference,
                           currency=None):
        """
        Refund against a previous ransaction
        """
        if currency is None:
            currency = settings.DATACASH_CURRENCY
        response = self.gateway.txn_refund(amount=amount,
                                           currency=currency,
                                           txn_reference=txn_reference)
        return self.handle_response(gateway.TXN_REFUND, order_number, amount,
                                    currency, response)

    def cancel_transaction(self, order_number, txn_reference):
        """
        Refund against a previous ransaction
        """
        response = self.gateway.cancel(txn_reference)
        return self.handle_response(
            gateway.CANCEL, order_number, None, None, response)

    # ========================
    # API - 1 stage processing
    # ========================

    def authorise(self, order_number, amount, bankcard=None, txn_reference=None,
                  billing_address=None, the3rdman_data=None, currency=None):
        """
        Debit a bankcard for the given amount

        A bankcard object or a txn_reference can be passed depending on whether
        you are using a new or existing bankcard.
        """
        if amount == 0:
            raise UnableToTakePayment("Order amount must be non-zero")
        if currency is None:
            currency = settings.DATACASH_CURRENCY

        merchant_ref = self.merchant_reference(order_number, gateway.AUTH)
        address_data = self.extract_address_data(billing_address)
        if bankcard:
            response = self.gateway.auth(card_number=bankcard.card_number,
                                         expiry_date=bankcard.expiry_date,
                                         amount=amount,
                                         currency=currency,
                                         merchant_reference=merchant_ref,
                                         ccv=bankcard.ccv,
                                         the3rdman_data=the3rdman_data,
                                         **address_data)
        elif txn_reference:
            response = self.gateway.auth(amount=amount,
                                         currency=currency,
                                         merchant_reference=merchant_ref,
                                         previous_txn_reference=txn_reference,
                                         the3rdman_data=the3rdman_data,
                                         **address_data)
        else:
            raise ValueError(
                "You must specify either a bankcard or a previous txn reference")

        return self.handle_response(gateway.AUTH, order_number, amount,
                                    currency, response)

    def refund(self, order_number, amount, bankcard=None, txn_reference=None,
               currency=None):
        """
        Return funds to a bankcard
        """
        if currency is None:
            currency = settings.DATACASH_CURRENCY
        merchant_ref = self.merchant_reference(order_number, gateway.REFUND)
        if bankcard:
            response = self.gateway.refund(card_number=bankcard.card_number,
                                           expiry_date=bankcard.expiry_date,
                                           amount=amount,
                                           currency=currency,
                                           merchant_reference=merchant_ref,
                                           ccv=bankcard.ccv)
        elif txn_reference:
            response = self.gateway.refund(
                amount=amount, currency=currency,
                merchant_reference=merchant_ref,
                previous_txn_reference=txn_reference)
        else:
            raise ValueError("You must specify either a bankcard or a previous txn reference")
        return self.handle_response(gateway.REFUND, order_number, amount,
                                    currency, response)



import logging

from django.views import generic
from django import http

from datacash import models

logger = logging.getLogger('datacash.the3rdman')


class CallbackView(generic.View):
    """
    Datacash will POST to this view when they have a fraud score
    for a transaction.  This view must respond with a simple string
    response within 1 second for it to be acknowledged.
    """

    def post(self, request, *args, **kwargs):
        # Create a fraud response object.  Other processes should listen
        # to the post create signal in order to hook fraud processing into
        # this order pipeline.
        try:
            # Datacash send both XML and query string with the same content
            # type header :( so we have to check for XML syntax.
            if '<?xml' in request.body:
                response = models.FraudResponse.create_from_xml(request.body)
            else:
                response = models.FraudResponse.create_from_querystring(request.body)
        except Exception, e:
            logger.error("Error raised handling response:\n%s", request.body)
            logger.exception(e)
            return http.HttpResponseServerError("error")
        else:
            logger.info("Successful response received with merchant ref %s",
                        response.merchant_order_ref)
            return http.HttpResponse("ok")
